from optimus.engines.spark.spark import Spark
from optimus.helpers.columns import parse_columns
from optimus.helpers.functions import path_is_local, prepare_path_local
from optimus.helpers.logger import logger

from optimus.helpers.types import *
from optimus.engines.base.io.save import BaseSave


class Save(BaseSave):

    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def json(self, path, mode="overwrite", encoding="UTF-8", num_partitions=1):

        df = self.root.data
        try:
            # na.fill enforce null value keys to the json output
            df.na.fill("") \
                .repartition(num_partitions) \
                .write \
                .option("encoding", encoding) \
                .format("json") \
                .mode(mode) \
                .save(path)
        except IOError as e:
            logger.print(e)
            raise

    def csv(self, path, header="true", mode="overwrite", single_file=True, storage_options=None, conn=None, sep=",",
            num_partitions=1):

        try:
            df = self.root

            if conn is not None:
                path = conn.path(path)
                storage_options = conn.storage_options

            try:
                if path_is_local(path):
                    prepare_path_local(path)
            except IOError as error:
                logger.print(error)
                raise

            columns = parse_columns(df, "*",
                                    filter_by_column_types=["date", "array", "vector", "binary", "null"])
            df = df.cols.cast(columns, "str").repartition(num_partitions)

            # Save to csv
            if single_file is True:
                print(path)
                # df.data.repartition(1).write.csv(path)
                df.data.toPandas().to_csv(path, header=True)
                # df.data.repartition(1).write.format('com.databricks.spark.csv').save(path,
                #                                                                 header='true')
            else:
                df.data.write.options(header=header).mode(mode).csv(path, sep=sep)

            # val conf    = sc.hadoopConfiguration
            # val src     = new Path(tmpFolder)
            # val fs      = src.getFileSystem(conf)
            # val oneFile = fs.listStatus(src).map(x => x.getPath.toString()).find(x => x.endsWith(format))
            # val srcFile = new Path(oneFile.getOrElse(""))
            # val dest    = new Path(filename)
            # fs.rename(srcFile, dest)
        except IOError as error:
            logger.print(error)
            raise

    def parquet(self, path, mode="overwrite", num_partitions=1):
        """
        Save data frame to a parquet file
        :param path: path where the spark will be saved.
        :param mode: Specifies the behavior of the save operation when data already exists.
                    "append": Append contents of this DataFrame to existing data.
                    "overwrite" (default case): Overwrite existing data.
                    "ignore": Silently ignore this operation if data already exists.
                    "error": Throw an exception if data already exists.
        :param num_partitions: the number of partitions of the DataFrame
        :return:
        """
        # This character are invalid as column names by parquet
        invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

        df = self.root

        def func(col_name):
            for i in invalid_character:
                col_name = col_name.replace(i, "_")
            return col_name

        df = df.cols.rename(func)

        columns = parse_columns(df, "*", filter_by_column_types=["null"])
        df = df.cols.cast(columns, "str")

        try:
            df.coalesce(num_partitions) \
                .write \
                .mode(mode) \
                .parquet(path)
        except IOError as e:
            logger.print(e)
            raise

    def avro(self, path, mode="overwrite", num_partitions=1):

        df = self.root
        try:
            avro_version = "avro"
            df.coalesce(num_partitions) \
                .write.format(avro_version) \
                .mode(mode) \
                .save(path)

        except IOError as e:
            logger.print(e)
            raise



import datetime
import numpy as np
from optimus.tests.base import TestBase
from optimus.helpers.json import json_encoding
from optimus.helpers.functions import deep_sort, df_dicts_equal, results_equal


def Timestamp(t):
    return datetime.datetime.strptime(t, "%Y-%m-%d %H:%M:%S")


NaT = np.datetime64('NaT')
nan = float("nan")
inf = float("inf")


class TestReplacePandas(TestBase):
    config = {'engine': 'pandas'}
    dict = {('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('Code', 'object'): ['123A', '456', 456, 'e', None, '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', 1, None, True, False], ('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}
    maxDiff = None

    def test_cols_replace_list_chars(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace(cols=['function'], search_by='chars', search=['True', 'False'], replace_by=['Maybe', 'Unlikely'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_full(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace(cols=['function'], search_by='full', search=['True', 'False'], replace_by=['Maybe', 'Unlikely'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value2_chars(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='chars', search=['leader', 'espionage', 'security'], replace_by='this is a test', ignore_case=True)
        expected = self.create_dataframe(data={('function', 'object'): ['this is a test', 'this is a test', 'this is a test', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value2_full(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='full', search=['leader', 'espionage', 'security'], replace_by='this is a test', ignore_case=True)
        expected = self.create_dataframe(data={('function', 'object'): ['this is a test', 'this is a test', 'this is a test', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value2_values(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='values', search=['leader', 'espionage', 'security'], replace_by='this is a test', ignore_case=True)
        expected = self.create_dataframe(data={('function', 'object'): ['this is a test', 'this is a test', 'this is a test', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value2_words(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='words', search=['leader', 'espionage', 'security'], replace_by='this is a test', ignore_case=True)
        expected = self.create_dataframe(data={('function', 'object'): ['this is a test', 'this is a test', 'this is a test', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value_chars(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='chars', search=['Leader', 'Firs', 'None', 'Secu'], replace_by='MATCH')
        expected = self.create_dataframe(data={('function', 'object'): ['MATCH', 'Espionage', 'MATCHrity', 'MATCHt Lieutenant', 'MATCH', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value_full(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='full', search=['Leader', 'Firs', 'None', 'Secu'], replace_by='MATCH')
        expected = self.create_dataframe(data={('function', 'object'): ['MATCH', 'Espionage', 'Security', 'First Lieutenant', 'MATCH', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value_values(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='values', search=['Leader', 'Firs', 'None', 'Secu'], replace_by='MATCH')
        expected = self.create_dataframe(data={('function', 'object'): ['MATCH', 'Espionage', 'Security', 'First Lieutenant', 'MATCH', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_value_words(self):
        df = self.df.copy().cols.select('function')
        result = df.cols.replace(cols='function', search_by='words', search=['Leader', 'Firs', 'None', 'Secu'], replace_by='MATCH')
        expected = self.create_dataframe(data={('function', 'object'): ['MATCH', 'Espionage', 'Security', 'First Lieutenant', 'MATCH', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_values(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace(cols=['function'], search_by='values', search=['True', 'False'], replace_by=['Maybe', 'Unlikely'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_list_words(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace(cols=['function'], search_by='words', search=['True', 'False'], replace_by=['Maybe', 'Unlikely'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_lists_list_chars(self):
        df = self.df.copy().cols.select(['function', 'japanese name'])
        result = df.cols.replace(cols=['function', 'japanese name'], search_by='chars', search=['atro', 'nochi'], replace_by=['Inochi', 'Megatron'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('japanese name', 'object'): ["['IMegatron', 'Convoy']", "['Bumble', 'Goldback']", "['Roadbuster']", "['Meister']", "['MegInochin']", "['Metroflex']"]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_lists_list_full(self):
        df = self.df.copy().cols.select(['function', 'Code', 'Cybertronian'])
        result = df.cols.replace(cols=['function', 'Code', 'Cybertronian'], search_by='full', search=['123', '456', '123A'], replace_by=['N/A', 'Number', 'String'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('Code', 'object'): ['String', 'Number', 456, 'e', None, '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', 1, None, True, False]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_lists_list_values(self):
        df = self.df.copy().cols.select(['function', 'japanese name'])
        result = df.cols.replace(cols=['function', 'japanese name'], search_by='values', search=['atro', 'nochi'], replace_by=['Inochi', 'Megatron'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_lists_list_words(self):
        df = self.df.copy().cols.select(['function', 'japanese name'])
        result = df.cols.replace(cols=['function', 'japanese name'], search_by='words', search=['atro', 'Megatron', 'nochi', 'Inochi'], replace_by=['--', 'a-match', '--', 'other-match'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('japanese name', 'object'): ["['other-match', 'Convoy']", "['Bumble', 'Goldback']", "['Roadbuster']", "['Meister']", "['a-match']", "['Metroflex']"]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_list_list_chars(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace_regex(cols=['function'], search_by='chars', search=['....', '.....', '......'], replace_by=['4c', '5c', '6c'])
        expected = self.create_dataframe(data={('function', 'object'): ['4cer', '5c', '4c4c', '5cc4c', '4c', '5ccon']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_list_list_full(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace_regex(cols=['function'], search_by='full', search=['....', '.....', '......'], replace_by=['4c', '5c', '6c'])
        expected = self.create_dataframe(data={('function', 'object'): ['6c', 'Espionage', 'Security', 'First Lieutenant', '4c', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_list_list_values(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace_regex(cols=['function'], search_by='values', search=['....', '.....', '......'], replace_by=['4c', '5c', '6c'])
        expected = self.create_dataframe(data={('function', 'object'): ['6c', 'Espionage', 'Security', 'First Lieutenant', '4c', 'Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_list_list_words(self):
        df = self.df.copy().cols.select(['function'])
        result = df.cols.replace_regex(cols=['function'], search_by='words', search=['....', '.....', '......'], replace_by=['4c', '5c', '6c'])
        expected = self.create_dataframe(data={('function', 'object'): ['6c', 'Espionage', 'Security', '5c Lieutenant', '4c', '6c Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_list_value_chars(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace_regex(cols='names', search_by='chars', search=['^a', '^e', '^i', '^o', '^u'], replace_by='starts with a vowel', ignore_case=True)
        expected = self.create_dataframe(data={('names', 'object'): ['starts with a vowelptimus', 'bumbl#ebéé  ', 'starts with a vowelronhide&', 'Jazz', 'Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_list_value_full(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace_regex(cols='names', search_by='full', search=['^a', '^e', '^i', '^o', '^u'], replace_by='starts with a vowel', ignore_case=True)
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_list_value_words(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace_regex(cols='names', search_by='words', search=['^a', '^e', '^i', '^o', '^u'], replace_by='starts with a vowel', ignore_case=True)
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_lists_list_chars(self):
        df = self.df.copy().cols.select(['function', 'Code', 'Cybertronian'])
        result = df.cols.replace_regex(cols=['function', 'Code', 'Cybertronian'], search_by='chars', search=[['\\d'], ['.*Tru.*', '.*Fal.*']], replace_by=['has a number', 'has boolean'])
        expected = self.create_dataframe(data={('function', 'object'): ['Lehas booleanhhas booleanshas booleanhas booleanhas booleannhas booleanmbehas booleanehas boolean', 'Espionhas booleange', 'Sechas booleanhas booleanity', 'has booleanihas booleansthas booleanLiehas booleantenhas booleannt', 'None', 'Bhas booleantthas booleanehas booleanSthas booleantion'], ('Code', 'object'): ['123A', '456', '456', 'e', 'None', '{cohas a numbere}'], ('Cybertronian', 'object'): ['has booleanhas booleanhas booleanse', 'has booleanhas booleanhas booleanehas boolean+has boolean1', '1', 'None', 'has booleanhas booleanhas booleane', 'has booleanhas booleanhas booleanse']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_lists_list_full(self):
        df = self.df.copy().cols.select(['function', 'Code', 'Cybertronian'])
        result = df.cols.replace_regex(cols=['function', 'Code', 'Cybertronian'], search_by='full', search=[['\\d', 'e'], ['.*True.*', '.*False.*']], replace_by=['is a number', 'is boolean'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('Code', 'object'): ['123A', '456', 456, 'is a number', None, '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', 1, None, True, False]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_lists_list_values(self):
        df = self.df.copy().cols.select(['function', 'Code', 'Cybertronian'])
        result = df.cols.replace_regex(cols=['function', 'Code', 'Cybertronian'], search_by='values', search=[['\\d'], ['.*True.*', '.*False.*']], replace_by=['is a number', 'is boolean'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('Code', 'object'): ['123A', '456', 456, 'is boolean', None, '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', 1, None, True, False]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_lists_list_words(self):
        df = self.df.copy().cols.select(['function', 'Code', 'Cybertronian'])
        result = df.cols.replace_regex(cols=['function', 'Code', 'Cybertronian'], search_by='words', search=[['\\d'], ['.*True.*', '.*False.*']], replace_by=['is a number', 'is boolean'])
        expected = self.create_dataframe(data={('function', 'object'): ['Leader', 'Espionage', 'Security', 'Firstis booleanLieutenant', 'None', 'Battleis booleanStation'], ('Code', 'object'): ['123A', '456', '456', 'is boolean', 'None', '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', '1', 'None', 'True', 'False']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_value_chars(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace_regex(cols='names', search_by='chars', search='.*atro.*', replace_by='must be Megatron')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'must be Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_value_full(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace_regex(cols='names', search_by='full', search='.*atro.*', replace_by='must be Megatron')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'must be Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_value_values(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace_regex(cols='names', search_by='values', search='.*atro.*', replace_by='must be Megatron')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'must be Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_regex_value_words(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace_regex(cols='names', search_by='words', search='.*atro.*', replace_by='must be Megatron')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'must be Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_types_chars(self):
        df = self.df.copy().cols.select('*')
        result = df.cols.replace(cols='*', search_by='chars', search=['at', 'an'], replace_by='this is a test')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megthis is a testron', 'Metroplex_)^$'], ('japanese name', 'object'): ["['Inochi', 'Convoy']", "['Bumble', 'Goldback']", "['Roadbuster']", "['Meister']", "['Megthis is a testron']", "['Metroflex']"], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenthis is a testt', 'None', 'Bthis is a testtle Stthis is a testion'], ('Code', 'object'): ['123A', '456', '456', 'e', 'None', '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', '1', 'None', 'True', 'False'], ('NullType', 'object'): ['None', 'None', 'None', 'None', 'None', 'None'], ('attributes', 'object'): ['[8.5344, 4300.0]', '[5.334, 2000.0]', '[7.9248, 4000.0]', '[3.9624, 1800.0]', '[None, 5700.0]', '[91.44, None]'], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): ["bytearray(b'Leader')", "bytearray(b'Espionage')", "bytearray(b'Security')", "bytearray(b'First Lieutenthis is a testt')", "bytearray(b'None')", "bytearray(b'Bthis is a testtle Stthis is a testion')"], ('height(ft)', 'object'): ['-28.0', '17.0', '26.0', '13.0', 'nthis is a test', '300.0'], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', 'None', 'None'], ('rank', 'object'): ['10', '7', '7', '8', '10', '8'], ('Date Type', 'object'): ['2016-09-10', '2015-08-10', '2014-06-24', '2013-06-24', '2012-05-10', '2011-04-10'], ('age', 'object'): ['5000000', '5000000', '5000000', '5000000', '5000000', '5000000'], ('timestamp', 'object'): ['2014-06-24', '2014-06-24', '2014-06-24', '2014-06-24', '2014-06-24', '2014-06-24'], ('weight(t)', 'object'): ['4.3', '2.0', '4.0', '1.8', '5.7', 'nthis is a test']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_types_full(self):
        df = self.df.copy().cols.select('*')
        result = df.cols.replace(cols='*', search_by='full', search=['at', 'an'], replace_by='this is a test')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('Code', 'object'): ['123A', '456', 456, 'e', None, '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', 1, None, True, False], ('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_types_values(self):
        df = self.df.copy().cols.select('*')
        result = df.cols.replace(cols='*', search_by='values', search=['at', 'an'], replace_by='this is a test')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('Code', 'object'): ['123A', '456', 456, 'e', None, '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', 1, None, True, False], ('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_types_words(self):
        df = self.df.copy().cols.select('*')
        result = df.cols.replace(cols='*', search_by='words', search=['at', 'an'], replace_by='this is a test')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('japanese name', 'object'): ["['Inochi', 'Convoy']", "['Bumble', 'Goldback']", "['Roadbuster']", "['Meister']", "['Megatron']", "['Metroflex']"], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('Code', 'object'): ['123A', '456', '456', 'e', 'None', '{code}'], ('Cybertronian', 'object'): ['False', 'True + 1', '1', 'None', 'True', 'False'], ('NullType', 'object'): ['None', 'None', 'None', 'None', 'None', 'None'], ('attributes', 'object'): ['[8.5344, 4300.0]', '[5.334, 2000.0]', '[7.9248, 4000.0]', '[3.9624, 1800.0]', '[None, 5700.0]', '[91.44, None]'], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): ["bytearray(b'Leader')", "bytearray(b'Espionage')", "bytearray(b'Security')", "bytearray(b'First Lieutenant')", "bytearray(b'None')", "bytearray(b'Battle Station')"], ('height(ft)', 'object'): ['-28.0', '17.0', '26.0', '13.0', 'nan', '300.0'], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', 'None', 'None'], ('rank', 'object'): ['10', '7', '7', '8', '10', '8'], ('Date Type', 'object'): ['2016-09-10', '2015-08-10', '2014-06-24', '2013-06-24', '2012-05-10', '2011-04-10'], ('age', 'object'): ['5000000', '5000000', '5000000', '5000000', '5000000', '5000000'], ('timestamp', 'object'): ['2014-06-24', '2014-06-24', '2014-06-24', '2014-06-24', '2014-06-24', '2014-06-24'], ('weight(t)', 'object'): ['4.3', '2.0', '4.0', '1.8', '5.7', 'nan']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_value_chars(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace(cols='names', search_by='chars', search='Optimus', replace_by='optimus prime')
        expected = self.create_dataframe(data={('names', 'object'): ['optimus prime', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_value_full(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace(cols='names', search_by='full', search='Optimus', replace_by='optimus prime')
        expected = self.create_dataframe(data={('names', 'object'): ['optimus prime', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_value_values(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace(cols='names', search_by='values', search='Optimus', replace_by='optimus prime')
        expected = self.create_dataframe(data={('names', 'object'): ['optimus prime', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_replace_value_words(self):
        df = self.df.copy().cols.select('names')
        result = df.cols.replace(cols='names', search_by='words', search='Optimus', replace_by='optimus prime')
        expected = self.create_dataframe(data={('names', 'object'): ['optimus prime', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))


class TestReplaceDask(TestReplacePandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestReplacePartitionDask(TestReplacePandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestReplaceCUDF(TestReplacePandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestReplaceDC(TestReplacePandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestReplacePartitionDC(TestReplacePandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}


try:
    import pyspark # pyright: reportMissingImports=false
except:
    pass
else:
    class TestReplaceSpark(TestReplacePandas):
        config = {'engine': 'spark'}


try:
    import vaex # pyright: reportMissingImports=false
except:
    pass
else:
    class TestReplaceVaex(TestReplacePandas):
        config = {'engine': 'vaex'}



from abc import ABC

import numpy as np
from sklearn.preprocessing import MinMaxScaler, MaxAbsScaler, StandardScaler

from optimus.engines.base.commons.functions import word_tokenize
from optimus.engines.base.functions import BaseFunctions


class DataFrameBaseFunctions(BaseFunctions, ABC):

    @staticmethod
    def map_partitions(dfd_or_series, func, *args, **kwargs):
        result = dfd_or_series.map_partitions(func, *args, **kwargs)
        result.index = dfd_or_series.index
        return result

    @staticmethod
    def dask_to_compatible(dfd):
        """Convert a Dask DataFrame to a compatible type of DataFrame
        Args:
            dfd (dd.DataFrame)
        """
        return dfd

    @staticmethod
    def all(series):
        return series.all()

    @staticmethod
    def not_all(series):
        return not series.all()

    @staticmethod
    def any(series):
        return series.any()

    @staticmethod
    def not_any(series):
        return not series.any()

    def reverse(self, series):
        return self.to_string(series).map(lambda v: v[::-1])

    def word_tokenize(self, series):
        return self.to_string(series).map(word_tokenize, na_action=None)

    def standard_scaler(self, series):
        return StandardScaler().fit_transform(self.to_float(series).values.reshape(-1, 1))

    def max_abs_scaler(self, series):
        return MaxAbsScaler().fit_transform(self.to_float(series).values.reshape(-1, 1))

    def min_max_scaler(self, series):
        return MinMaxScaler().fit_transform(self.to_float(series).values.reshape(-1, 1))

    @staticmethod
    def heatmap(df, bins):
        heatmap, xedges, yedges = np.histogram2d(df['x'].values, df['y'].values, bins=bins)
        return heatmap, xedges, yedges



from optimus.helpers.core import val_to_list


class RaiseIt:

    @staticmethod
    def _or(values):
        values = val_to_list(values)
        if len(values) == 1:
            return f"'{values[0]}'"
        return ", ".join([f"'{v}'" for v in values[0:-1]]) + " or " + f"'{values[-1]}'"

    @staticmethod
    def _and(values):
        values = val_to_list(values)
        if len(values) == 1:
            return f"'{values[0]}'"
        return ", ".join([f"'{v}'" for v in values[0:-1]]) + " and " + f"'{values[-1]}'"

    @classmethod
    def type_error(cls, vars, data_types):
        """
        Raise a TypeError exception
        :param var:
        :param data_types: data types expected as strings
        :return:
        """
        from optimus.helpers.debug import get_var_name

        if get_var_name(vars):
            vars = (vars,)

        var_names = []
        var_types = []

        for arg in vars:
            var_names.append(get_var_name(arg))
            var_types.append(type(arg))

        var_names = cls._and(var_names)
        var_types = cls._and(var_types)
        data_types = cls._or(data_types)

        raise TypeError(f"{var_names} must be of type {data_types}, received {var_types}")

    @staticmethod
    def length_error(var1: list, var2: (list, int)) -> Exception:
        """
        Raise a ValueError exception
        :param var1: variable to check for length
        :param var2: variable or integer to compare with var1
        :return:
        """
        from optimus.helpers.debug import get_var_name

        if isinstance(var2, int):
            length_var2 = str(var2)
        else:
            length_var2 = str(len(var2))

        raise ValueError("'{var2_name}' must be length '{var1_length}', received '{var2_length}'"
                         .format(var2_name=get_var_name(var2), var1_length=str(len(var1)), var2_length=length_var2))

    @staticmethod
    def not_ready_error(message):
        raise NotReady(message)

    @classmethod
    def value_error(cls, vars, data_values=None, extra_text=""):
        """
        Raise a ValueError exception
        :param var:
        :param data_values: values accepted by the variable. str/list
        :param extra_text: Additional final info about the error
        :return:
        """
        from optimus.helpers.debug import get_var_name

        if get_var_name(vars):
            vars = (vars,)

        var_names = []
        var_values = []

        for arg in vars:
            var_names.append(get_var_name(arg))
            var_values.append(arg)

        var_names = cls._and(var_names)
        var_values = cls._and(var_values)

        if data_values:
            data_values = cls._or(val_to_list(data_values))
            err = f"{var_names} must be {data_values}, received {var_values}."
        else:
            err = f"Invalid {var_names}: {var_values}."

        if extra_text:
            err += f" {extra_text}."

        raise ValueError(err)

    @staticmethod
    def type(cls, var):
        """
        Raise and exception of the type specified
        :param cls: Exception Class to be Raised
        :param var:
        :return:
        """

        from optimus.helpers.debug import get_var_name
        raise cls("'{var_name}' error".format(var_name=get_var_name(var), var_type=var))

    @staticmethod
    def message(cls, message):
        """
        Raise and exception of the type specified with a message
        :param cls: Exception Class to be Raised
        :param message:
        :return:
        """
        raise cls(message)


# define Python user-defined exceptions
class Error(Exception):
    """Base class for other exceptions"""
    pass


class NotReady(Error):
    pass



import tarfile

import glob

from optimus.helpers.functions import prepare_path


class Extract:
    @staticmethod
    def gz(path, *args, **kwargs):
        """
        Loads a dataframe from a gz file.
        :param path: path or location of the file. Must be string dataType
        :param args: custom argument to be passed to the internal function
        :param kwargs: custom keyword arguments to be passed to the internal function
        :return: Spark Dataframe
        """
        file, file_name = prepare_path(path, "gz")

        import gzip
        import shutil
        with gzip.open(file, 'rb') as f_in:
            print(f_in)
            with open('file.txt', 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

        print(file, file_name)

    @staticmethod
    def tar(input_file, output_path="", *args, **kwargs):
        """
        Extract .tar or .tar.gz files
        :param input_file:
        :param output_path:
        :param args:
        :param kwargs:
        :return: List of file names
        """

        if input_file.endswith("tar.gz"):
            mode = "r:gz"
        elif input_file.endswith("tar"):
            mode = "r:"

        tar = tarfile.open(input_file, mode)
        tar.extractall(path=output_path)
        tar.close()

        # filename
        output_path = input_file.split('.')[0]
        return glob.glob(output_path)



from abc import abstractmethod

import numpy as np

from optimus import Optimus
from optimus.engines.base.constants import BaseConstants
from optimus.engines.base.functions import BaseFunctions
from optimus.engines.base.io.connect import Connect
from optimus.engines.dask.io.jdbc import JDBC
from optimus.helpers.logger import logger


class BaseEngine:

    @staticmethod
    def verbose(verbose):
        """
        Enable verbose mode
        :param verbose:
        :return:
        """

        logger.active(verbose)

    def __getstate__(self):
        return self.engine

    def __setstate__(self, state):
        return Optimus(state)


    @staticmethod
    def connect(driver=None, host=None, database=None, user=None, password=None, port=None, schema="public",
                oracle_tns=None, oracle_service_name=None, oracle_sid=None, presto_catalog=None,
                cassandra_keyspace=None, cassandra_table=None, bigquery_project=None, bigquery_dataset=None):
        """
        Create the JDBC string connection
        :return: JDBC object
        """

        return JDBC(host, database, user, password, port, driver, schema, oracle_tns, oracle_service_name, oracle_sid,
                    presto_catalog, cassandra_keyspace, cassandra_table, bigquery_project, bigquery_dataset)

    @property
    def nan(self):
        """
        Create the JDBC string connection
        :return: JDBC object
        """
        return np.nan

    @abstractmethod
    def create(self):
        pass

    @property
    def constants(self):
        return BaseConstants()

    @property
    def F(self):
        return BaseFunctions()

    @abstractmethod
    def load(self):
        pass

    @abstractmethod
    def engine(self):
        pass

    @property
    def engine_label(self):
        return self.__class__.__name__

    @abstractmethod
    def remote_run(self, callback, *args, **kwargs):
        pass

    @abstractmethod
    def remote_submit(self, callback, *args, **kwargs):
        pass

    @abstractmethod
    def submit(self, func, *args, priority=0, pure=False, **kwargs):
        pass



import re
from abc import ABC

import cudf

from optimus.engines.base.functions import BaseFunctions
from optimus.helpers.core import val_to_list


class CUDFBaseFunctions(BaseFunctions, ABC):
    @staticmethod
    def to_dict(series) -> dict:
        series.name = str(series.name)
        return series.to_pandas().to_dict()

    def to_items(self, series) -> dict:
        """
        Convert series to a list of tuples [(index, value), ...]
        """
        df = series.reset_index()
        return df.to_pandas().to_dict(orient='split')['data']

    def is_integer(self, series):
        if str(series.dtype) in self.constants.DATETIME_INTERNAL_TYPES:
            return cudf.Series([False] * len(series))
        if str(series.dtype) in self.constants.INT_INTERNAL_TYPES:
            return cudf.Series([True] * len(series))
        return series.astype(str).str.isinteger()

    def is_float(self, series):
        if str(series.dtype) in self.constants.DATETIME_INTERNAL_TYPES:
            return cudf.Series([False] * len(series))
        return series.astype(str).str.isfloat()

    def is_numeric(self, series):
        if str(series.dtype) in self.constants.DATETIME_INTERNAL_TYPES:
            return cudf.Series([False] * len(series))
        return series.astype(str).str.isfloat()

    def is_string(self, series):
        if str(series.dtype) in self.constants.STRING_INTERNAL_TYPES:
            return cudf.Series([True] * len(series))
        return cudf.Series([False] * len(series))

    def _to_integer(self, series, default=0):
        return cudf.to_numeric(series, errors="coerce", downcast="integer")

    def _to_float(self, series):
        import numpy as np
        # Workaround to https://github.com/rapidsai/cudf/issues/10049
        if series.dtype.type == np.bool_:
            return series.astype(float)
        else:
            return cudf.to_numeric(series, errors="coerce", downcast="float")

    @staticmethod
    def to_string(series):
        return series.astype(str)

    def _to_datetime(self, value, format):
        return cudf.to_datetime(value, format=format, errors="coerce")

    def replace_chars(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)%s' % re.escape(s) for s in search]
        else:
            str_regex = [r'%s' % re.escape(s) for s in search]
        return self.to_string(series).str.replace(str_regex, replace_by, regex=True)

    def contains(self, series, value, case, flags, na, regex):
        return self.to_string_accessor(series).contains(value, case=case, flags=flags, regex=regex).fillna(na)



from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.pandas.dataframe import PandasDataFrame


class IbisDataFrame(BaseDataFrame):

    def _base_to_dfd(self, pdf, n_partitions) -> 'InternalDataFrameType':
        pass

    @property
    def save(self) -> 'BaseSave':
        pass

    def visualize(self):
        pass

    def stratified_sample(self, col_name, seed: int = 1) -> 'DataFrameType':
        pass

    def _iloc(self, lower_bound, upper_bound, copy=True):
        dfd = self.data[lower_bound: upper_bound]
        if copy:
            dfd = dfd.reset_index(drop=True)
        return self.root.new(dfd, meta=self.root.meta)

    def graph(self) -> dict:
        pass

    @property
    def rows(self):
        from optimus.engines.ibis.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.ibis.columns import Cols
        return Cols(self)

    @property
    def functions(self):
        from optimus.engines.ibis.functions import IbisFunctions
        return IbisFunctions(self)

    @staticmethod
    def sample(n=10, random=False):
        pass

    @staticmethod
    def melt(id_vars, value_vars, var_name="variable", value_name="value", data_type="str"):
        pass

    @staticmethod
    def query(sql_expression):
        pass

    @staticmethod
    def partitions():
        pass

    @staticmethod
    def partitioner():
        pass

    def repartition(self, n=None, *args, **kwargs):
        return self.root

    @staticmethod
    def show():
        pass

    @staticmethod
    def debug():
        pass

    def compile(self):
        # return str(ibis.impala.compiler(self.parent.data))
        return str(self.root.data.compile())

    def to_pandas(self):
        return self.root.data.execute()

    def to_optimus_dataframe(self):
        return

    def to_optimus_pandas(self):
        return PandasDataFrame(self.root.data.execute(), op=self.op)


# Helpers
from optimus.engines.base.dataframe.rows import DataFrameBaseRows
from optimus.engines.base.pandas.rows import PandasBaseRows
from optimus.engines.base.rows import BaseRows


# from optimus.infer_spark import is_list_of_spark_dataframes


class Rows(DataFrameBaseRows, PandasBaseRows, BaseRows):
    pass



from optimus.engines.base.io.save import BaseSave, DEFAULT_MODE
from optimus.helpers.logger import logger

from optimus.helpers.types import *
from optimus.engines.base.io.save import BaseSave


class Save(BaseSave):
    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def json(self, path, *args, **kwargs):

        df = self.root.data
        try:
            df.to_json(filename=path, *args, **kwargs)

        except IOError as e:
            logger.print(e)
            raise

    def csv(self, path, mode=DEFAULT_MODE, *args, **kwargs):

        try:
            df = self.root.data
            # columns = parse_columns(self, "*",
            #                         filter_by_column_types=["date", "array", "vector", "binary", "null"])
            # df = df.cols.cast(columns, "str").repartition(num_partitions)

            # Dask reference
            # https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_csv
            # df.to_csv(filename=path)
            df.to_csv(filename=path, mode=mode, *args, **kwargs)

        except IOError as error:
            logger.print(error)
            raise

    def parquet(self, path, mode=DEFAULT_MODE, num_partitions=1, compression="snappy", *args, **kwargs):

        # This character are invalid as column names by parquet
        invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

        def func(col_name):
            for i in invalid_character:
                col_name = col_name.replace(i, "_")
            return col_name

        df = self.root.cols.rename(func)

        try:
            df.to_parquet(path, compression=compression)
            print(path)
        except IOError as e:
            logger.print(e)
            raise

    def avro(self, path, *args, **kwargs):
        raise NotImplementedError('Not implemented yet')



import databricks.koalas as ks
import numpy as np
# These function can return a Column Expression or a list of columns expression
# Must return None if the data type can not be handle
import pandas as pd
from pyspark.sql import functions as F

from optimus.engines.base.commons.functions import word_tokenize
from optimus.engines.base.dataframe.functions import DataFrameBaseFunctions
from optimus.engines.base.pandas.functions import PandasBaseFunctions
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.raiseit import RaiseIt


class SparkFunctions(PandasBaseFunctions, DataFrameBaseFunctions):
    _engine = ks

    def _to_float(self, series):
        """
        Converts a series values to floats
        """
        return series.astype("float")

    def _to_integer(self, series, default=0):
        """
        Converts a series values to integers
        """
        return series.astype("integer")

    def _to_string(self, series):
        """
        Converts a series values to strings
        """
        return series.astype("str")

    def _to_boolean(self, series):
        """
        Converts a series values to bool
        """
        return series.astype("bool")

    def hist(col_name, df, buckets, min_max=None, dtype=None):
        """
        Create a columns expression to calculate a column histogram
        :param col_name:
        :param df:
        :param buckets:
        :param min_max: Min and max value necessary to calculate the buckets
        :param dtype: Column datatype to calculate the related histogram. Int, String and Dates return different histograms
        :return:
        """
        PYSPARK_NUMERIC_TYPES = ["byte", "short", "big", "int", "double", "float"]
        def is_column_a(df, column, dtypes):
            """
            Check if column match a list of data types
            :param df: dataframe
            :param column: column to be compared with
            :param dtypes: types to be checked
            :return:
            """
            column = val_to_list(column)

            if len(column) > 1:
                RaiseIt.length_error(column, 1)

            data_type = tuple(val_to_list(parse_spark_dtypes(dtypes)))
            column = one_list_to_val(column)

            # Filter columns by data type
            return isinstance(df.schema[column].dataType, data_type)

        def create_exprs(_input_col, _buckets, _func):
            def count_exprs(_exprs):
                return F.sum(F.when(_exprs, 1).otherwise(0))

            _exprs = []
            for i, b in enumerate(_buckets):
                lower = b["lower"]
                upper = b["upper"]

                if is_numeric(lower):
                    lower = round(lower, 2)

                if is_numeric(upper):
                    upper = round(upper, 2)

                if len(_buckets) == 1:
                    count = count_exprs(
                        (_func(_input_col) == lower))
                else:
                    if i == len(_buckets):
                        count = count_exprs(
                            (_func(_input_col) > lower) & (_func(_input_col) <= upper))
                    else:
                        count = count_exprs(
                            (_func(_input_col) >= lower) & (_func(_input_col) < upper))

                info = F.create_map(F.lit("count"), count.cast("int"), F.lit("lower"), F.lit(lower), F.lit("upper"),
                                    F.lit(upper)).alias(
                    "hist_agg" + "_" + _input_col + "_" + str(b["bucket"]))
                _exprs.append(info)
            _exprs = F.array(*_exprs).alias("hist" + _input_col)
            return _exprs

        def hist_numeric(_min_max, _buckets):
            if _min_max is None:
                _min_max = df.agg(F.min(col_name).alias("min"), F.max(col_name).alias("max")).to_dict()[0]

            if _min_max["min"] is not None and _min_max["max"] is not None:
                _buckets = create_buckets(_min_max["min"], _min_max["max"], _buckets)
                _exprs = create_exprs(col_name, _buckets, F.col)
            else:
                _exprs = None

            return _exprs

        def hist_string(_buckets):
            _buckets = create_buckets(0, 50, _buckets)
            func = F.length
            return create_exprs(col_name, _buckets, func)

        def hist_date():
            now = datetime.datetime.now()
            current_year = now.year
            oldest_year = 1950

            # Year
            _buckets = create_buckets(oldest_year, current_year, current_year - oldest_year)
            func = F.year
            year = create_exprs(col_name, _buckets, func)

            # Month
            _buckets = create_buckets(1, 12, 11)
            func = F.month
            month = create_exprs(col_name, _buckets, func)

            # Day
            _buckets = create_buckets(1, 31, 31)
            func = F.dayofweek
            day = create_exprs(col_name, _buckets, func)

            # Hour
            _buckets = create_buckets(0, 23, 23)
            func = F.hour
            hour = create_exprs(col_name, _buckets, func)

            # Min
            _buckets = create_buckets(0, 60, 60)
            func = F.minute
            minutes = create_exprs(col_name, _buckets, func)

            # Second
            _buckets = create_buckets(0, 60, 60)
            func = F.second
            second = create_exprs(col_name, _buckets, func)

            exprs = F.create_map(F.lit("years"), year, F.lit("months"), month, F.lit("weekdays"), day,
                                 F.lit("hours"), hour, F.lit("minutes"), minutes, F.lit("seconds"), second)

            return exprs

        if dtype is not None:
            col_dtype = dtype[col_name]["dtype"]
            if col_dtype == "int" or col_dtype == "decimal":
                exprs = hist_numeric(min_max, buckets)
            elif col_dtype == "string":
                exprs = hist_string(buckets)
            elif col_dtype == "date":
                exprs = hist_date()
            else:
                exprs = None
        else:
            if is_column_a(df, col_name, PYSPARK_NUMERIC_TYPES):
                exprs = hist_numeric(min_max, buckets)

            elif is_column_a(df, col_name, "str"):
                exprs = hist_string(buckets)

            elif is_column_a(df, col_name, "date") or is_column_a(df, col_name, "timestamp"):
                exprs = hist_date()
            else:
                exprs = None

        return exprs

    def create_exprs(_input_col, _buckets, _func):
        def count_exprs(_exprs):
            return F.sum(F.when(_exprs, 1).otherwise(0))

        _exprs = []
        for i, b in enumerate(_buckets):
            lower = b["lower"]
            upper = b["upper"]

            if is_numeric(lower):
                lower = round(lower, 2)

            if is_numeric(upper):
                upper = round(upper, 2)

            if len(_buckets) == 1:
                count = count_exprs(
                    (_func(_input_col) == lower))
            else:
                if i == len(_buckets):
                    count = count_exprs(
                        (_func(_input_col) > lower) & (_func(_input_col) <= upper))
                else:
                    count = count_exprs(
                        (_func(_input_col) >= lower) & (_func(_input_col) < upper))

            info = F.create_map(F.lit("count"), count.cast("int"), F.lit("lower"), F.lit(lower), F.lit("upper"),
                                F.lit(upper)).alias(
                "hist_agg" + "_" + _input_col + "_" + str(b["bucket"]))
            _exprs.append(info)
        _exprs = F.array(*_exprs).alias("hist" + _input_col)
        return _exprs

    @staticmethod
    def dask_to_compatible(dfd):
        from optimus.helpers.converter import dask_dataframe_to_pandas
        return ks.from_pandas(dask_dataframe_to_pandas(dfd))

    @staticmethod
    def new_df(*args, **kwargs):
        return ks.from_pandas(pd.DataFrame(*args, **kwargs))

    @staticmethod
    def df_concat(df_list):
        return ks.concat(df_list, axis=0, ignore_index=True)

    def word_tokenize(self, series):
        return self.to_string(series).map(word_tokenize, na_action=None)

    def count_zeros(self, series, *args):
        return int((self.to_float(series).values == 0).sum())

    def kurtosis(self, series):
        return self.to_float(series).kurtosis()

    def skew(self, series):
        return self.to_float(series).skew()

    def exp(self, series):
        return np.exp(self.to_float(series))

    def sqrt(self, series):
        return np.sqrt(self.to_float(series))

    def reciprocal(self, series):
        return np.reciprocal(self.to_float(series))

    def radians(self, series):
        return np.radians(self.to_float(series))

    def degrees(self, series):
        return np.degrees(self.to_float(series))

    def ln(self, series):
        return np.log(self.to_float(series))

    def log(self, series, base=10):
        return np.log(self.to_float(series)) / np.log(base)

    def sin(self, series):
        return np.sin(self.to_float(series))

    def cos(self, series):
        return np.cos(self.to_float(series))

    def tan(self, series):
        return np.tan(self.to_float(series))

    def asin(self, series):
        return np.arcsin(self.to_float(series))

    def acos(self, series):
        return np.arccos(self.to_float(series))

    def atan(self, series):
        return np.arctan(self.to_float(series))

    def sinh(self, series):
        return np.arcsinh(self.to_float(series))

    def cosh(self, series):
        return np.cosh(self.to_float(series))

    def tanh(self, series):
        return np.tanh(self.to_float(series))

    def asinh(self, series):
        return np.arcsinh(self.to_float(series))

    def acosh(self, series):
        return np.arccosh(self.to_float(series))

    def atanh(self, series):
        return np.arctanh(self.to_float(series))

    def floor(self, series):
        return np.floor(self.to_float(series))

    def ceil(self, series):
        return np.ceil(self.to_float(series))

    def normalize_chars(self, series):
        return series.str.normalize("NFKD").str.encode('ascii', errors='ignore').str.decode('utf8')

    def format_date(self, series, current_format=None, output_format=None):
        return ks.to_datetime(series, format=current_format,
                              errors="coerce").dt.strftime(output_format).reset_index(drop=True)

    def time_between(self, series, value=None, date_format=None):

        value_date_format = date_format

        if is_list_or_tuple(date_format) and len(date_format) == 2:
            date_format, value_date_format = date_format

        if is_list_or_tuple(value) and len(value) == 2:
            value, value_date_format = value

        date = pd.to_datetime(series, format=date_format, errors="coerce")
        value = pd.Timestamp.now() if value is None else pd.to_datetime(value, format=value_date_format,
                                                                        errors="coerce")

        return (value - date)



import glob

import pandas as pd
import json
from glom import glom

import optimus.helpers.functions_spark
from optimus.infer import is_dict, is_list_value, is_str, is_int

META = "_meta"
PROPERTIES = "_properties"
ITEMS = "_items"

COL_DEPTH = "depth"


class JSON:
    def __init__(self):
        self.data = None

    def load(self, path):
        """
        Load a file in JSON  format
        :param path:
        :return:
        """
        all_json = glob.glob(path, recursive=True)
        # pd.read_json("data/corona.json")
        with open(all_json[0]) as f:
            self.data = json.load(f)

    def schema(self) -> dict:
        """
        Return a dict with the count, data_type and nested structure
        :return:
        """

        def _schema(_data, _keys):
            if isinstance(_data, dict):
                for x, y in _data.items():
                    if is_dict(y):
                        _keys[x] = {
                            META: {"count": len(y), "data_type": type(y)}}
                        if len(y) > 0:
                            _keys[x][PROPERTIES] = {}
                            _schema(y, _keys[x][PROPERTIES])
                    elif is_list_value(y):
                        _keys[x] = {
                            META: {"count": len(y), "data_type": type(y)}}
                        if len(y) > 0:
                            _keys[x] = {ITEMS: {PROPERTIES: {}, META: {
                                "count": len(y), "data_type": type(y)}}}
                            _schema(y, _keys[x][ITEMS][PROPERTIES])
                    elif is_str(y):
                        _keys[x] = {
                            META: {"count": len(y), "data_type": type(y)}}
                        _schema(y, _keys[x])
                    elif is_int(y):
                        _keys[x] = {META: {"data_type": type(y)}}
                        _schema(y, _keys[x])

            elif is_list_value(_data):
                for x in _data:
                    _schema(x, _keys)

        keys = {}
        _schema(self.data, keys)
        return keys

    def freq(self, n: int = 100):
        """
        Calculate the count on every dict or list in the json
        :param n:
        :return:
        """

        def _profile(keys, parent, result=None):
            for key, values in keys.items():
                if values.get(PROPERTIES):
                    _meta = values.get(META)
                    _properties = values.get(PROPERTIES)
                elif values.get(ITEMS):
                    _meta = values.get(ITEMS).get(META)
                    _properties = values.get(ITEMS).get(PROPERTIES)

                if values.get(PROPERTIES) or values.get(ITEMS):
                    optimus.helpers.functions_spark.append(
                        [key, _meta["count"], _meta["data_type"], parent, len(parent)])
                    _profile(_properties, parent + [key], result=result)

        data = []
        _profile(self.schema(), [], data)
        pdf = pd.DataFrame(
            data, columns=['key', 'count', 'data_type', 'path', COL_DEPTH])
        return pdf.sort_values(by=["count", COL_DEPTH], ascending=[False, True]).rows.limit(n).to_dict(orient='row')

    def flatten(self, path):
        """
        Flatten a dict from a dict path
        :param path:
        :return:
        """

        def _flatten_json(_values):
            out = {}

            def flatten(x, name=''):
                if type(x) is dict:
                    for a in x:
                        flatten(x[a], name + a + '_')
                elif type(x) is list:
                    # i = 0
                    for a in x:
                        # flatten(a, name + str(i) + '_')
                        flatten(a, name + '_')
                        # i += 1
                else:
                    out[name[:-1]] = x

            flatten(_values)
            return out

        result = []
        value = glom(self.data, path, skip_exc=KeyError)
        if is_list_value(value):
            for i in value:
                result.append((_flatten_json(i)))
        elif is_dict(value):
            for i, j in value.items():
                a = {path: i}
                a.update(_flatten_json(j))
                result.append(a)
        return result

    def to_pandas(self, path):
        result = self.flatten(path)
        return pd.DataFrame(data=result)



import warnings

import pandavro as pdx

from optimus.engines.base.io.save import BaseSave
from optimus.helpers.logger import logger

from optimus.helpers.types import *
from optimus.engines.base.io.save import BaseSave


class Save(BaseSave):
    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def json(self, path, mode="w", *args, **kwargs):

        df = self.root.data
        try:
            df.to_json(path, mode=mode, *args, **kwargs)

        except IOError as e:
            logger.print(e)
            raise

    def csv(self, path, mode="rb", *args, **kwargs):
        try:
            dfd = self.root.cols.cast("*", "str").data
            dfd.to_csv(path, index=False, mode=mode, *args, **kwargs)

        except IOError as error:
            logger.print(error)
            raise

    def parquet(self, path, mode="overwrite", num_partitions=1, *args, **kwargs):

        # This character are invalid as column names by parquet
        invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

        def func(col_name):
            for i in invalid_character:
                col_name = col_name.replace(i, "_")
            return col_name

        df = self.root.cols.rename(func)

        try:
            df.data.to_parquet(path, mod=mode, numpartitions=num_partitions)
        except IOError as e:
            logger.print(e)
            raise

    def excel(self, path, **kwargs):

        try:
            # df = self.root.data
            # columns = parse_columns(self, "*",
            #                         filter_by_column_types=["date", "array", "vector", "binary", "null"])
            df = self.root.cols.cast("*", "str")

            # Dask reference
            # https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_csv
            df.to_pandas().to_excel(path, index=False)

        except IOError as error:
            logger.print(error)
            raise

    def orc(self, path, **kwargs):
        try:
            df = self.root.data
            df.to_orc(path, index=False, **kwargs)

        except IOError as error:
            logger.print(error)
            raise

    def avro(self, path, **kwargs):
        warnings.warn("Using CPU via pandavro to read the avro dataset")
        pdx.to_avro(path, self.root.to_pandas())



import errno
from optimus.helpers.core import val_to_list
import warnings
import os
from io import UnsupportedOperation
from pprint import pformat

from optimus.engines.base.basedataframe import BaseDataFrame

from optimus.infer import is_str, is_function, is_list_empty
from optimus.helpers.logger import logger


class TestCreator:

    def __init__(self, op=None, df=None, name=None, path="", create_path="..", configs={}, **kwargs):
        """
        Create python code with unit test functions for Optimus.
        :param op: optimus instance
        :param df: Spark Dataframe
        :param name: Name of the Test Class
        :type path: folder where tests will be written individually. run() nneds to be runned to merge all the tests.
        :param options: dictionary with configuration options
            import: Libraries to be added. 
        :param configs: Array of dictionaries with:
            engine: Engine to use. 
            n_partitions: Number of partitions of created dataframes (if supported)

        """
        if path and len(path):
            create_path += f"/{path}"

        self.op = op
        self.df = df
        self.name = name
        self.path = path
        self.create_path = create_path
        self.options = kwargs
        self.configs = configs
        self.created = []

    def run(self, clear=True):
        """
        Return the tests in text format
        :return:
        """

        filename = self.create_path + "/" + "test_created__" + self.name + ".py"

        test_file = open(filename, 'w', encoding='utf-8')
        print("Creating file " + filename)

        # Imports
        _imports = [
            "import datetime",
            "import numpy as np",
            "from optimus.tests.base import TestBase",
            "from optimus.helpers.json import json_encoding",
            "from optimus.helpers.functions import deep_sort, df_dicts_equal, results_equal",
            "\n\ndef Timestamp(t):\n    return datetime.datetime.strptime(t, \"%Y-%m-%d %H:%M:%S\")\n\n",
            "NaT = np.datetime64('NaT')",
            "nan = float(\"nan\")",
            "inf = float(\"inf\")",
        ]

        if self.options.get("imports", None) is not None:
            for i in self.options["imports"]:
                _imports.append(i)
        for i in _imports:
            test_file.write(i + "\n")

        classes = {}

        for name, config in self.configs.items():
            key = "".join(w.title() for w in self.name.split("_")) + name
            classes.update({key: config})

        # Class name
        base_class = "Test"+list(classes.keys())[0]
        cls = "\n\nclass " + base_class + "(TestBase):\n"
        test_file.write(cls)

        # First Config
        test_file.write("    config = " +
                        pformat(list(classes.values())[0], compact=True) + "\n")

        # Global Dataframe
        if self.df is not None:
            test_file.write("    dict = " + self.df.export(data_types="internal") + "\n")

        test_file.write("    maxDiff = None\n")

        for root, dirs, files in os.walk(self.create_path):
            files.sort()
            for file in files:
                if file.endswith(".test"):
                    full_path = os.path.join(root, file)

                    with open(full_path, 'r', encoding='utf-8') as opened_file:
                        try:
                            text = opened_file.read()

                            test_file.write(text)
                            opened_file.close()
                        except UnsupportedOperation:
                            print("file seems to be empty")

        for name, config in list(classes.items())[1:]:

            class_config = config.copy()

            try_package = class_config.pop("try_package", None)

            test_file.write("\n\n")
            extra_tab = ""
            if try_package:
                extra_tab = "    "
                test_file.write("try:\n")
                try_package = val_to_list(try_package)
                for package in try_package:
                    test_file.write(f"    import {package} # pyright: reportMissingImports=false\n")
                test_file.write("except:\n")
                test_file.write("    pass\n")
                test_file.write("else:\n")
                

            test_file.write(f"{extra_tab}class Test{name}({base_class}):\n")
            test_file.write(f"{extra_tab}    config = {pformat(class_config, compact=True)}\n")

        test_file.close()

        if clear:
            self.clear()

        print("Done")

    def clear(self):
        for method, variant in self.created:
            self.delete(method, variant)

    def create(self, method=None, variant=None, df=None, compare_by="df", select_cols=False, additional_method=[], args=(), **kwargs):
        """
        This is a helper function that output python tests for Spark DataFrames.
        :param method: Method to be tested
        :param variant: The test name will be create using the method param. This will be added as a suffix in case you want to customize the test name.
        :param df: Object to be tested
        :param compare_by: 'df', 'json' or 'dict'
        :param additional_method:
        :param args: Arguments to be used in the method
        :param kwargs: Keyword arguments to be used in the functions
        :return:
        """

        buffer = []

        def add_buffer(value, tab=1):
            buffer.append(("    " * tab) + value)

        # Create name
        name = []

        if method is not None:
            name.append(method.replace(".", "_"))

        additional_method = val_to_list(additional_method)

        for a_method in additional_method:
            name.append(a_method)

        if variant is not None:
            name.append(variant)

        test_name = "_".join(name)

        func_test_name = "test_" + test_name + "()"

        print(f"Creating {func_test_name} test function...")
        logger.print(func_test_name)

        func_test_name = "test_" + test_name + "(self)"

        filename = test_name + ".test"

        add_buffer("\n", 0)
        add_buffer("def " + func_test_name + ":\n")

        if select_cols is True:
            select_cols = kwargs["cols"] if "cols" in kwargs else args[0] if len(
                args) else False

        if df is None and self.df is None and (len(args)+len(kwargs)):
            df = self.op.create.dataframe(*args, **kwargs)
            df_func = df

        if df is None:
            # Use the main df
            df = self.df
            if select_cols:
                df = df.cols.select(select_cols)
                add_buffer(
                    f"df = self.df.copy().cols.select({pformat(select_cols, compact=True)})\n", 2)
            else:
                add_buffer("df = self.df.copy()\n", 2)
            df_func = df
        elif isinstance(df, (BaseDataFrame,)):
            if select_cols:
                df = df.cols.select(select_cols)
            add_buffer("df = self.create_dataframe(data=" + df.export(data_types="internal") + ", force_data_types=True)\n", 2)
            df_func = df
        else:
            if select_cols:
                df = [df[col] for col in df if df in select_cols] if select_cols != "*" else df
            add_buffer("df = self.create_dataframe(data=" + pformat(df, compact=True, sort_dicts=False) +
                       ", force_data_types=True)\n", 2)
            df_func = df

        # Process simple arguments
        _args = []
        for v in args:
            if is_function(v):
                _args.append(v.__qualname__)
            elif isinstance(v, (BaseDataFrame,)):
                _df = "self.create_dataframe(data=" + v.export(data_types="internal") + ", force_data_types=True)"
                _args.append(_df)
            elif isinstance(v, list) and isinstance(v[0], (BaseDataFrame,)):
                _dfs = ["self.create_dataframe(data=" + dv.export(data_types="internal") + ", force_data_types=True)" for dv in v]
                _dfs = "[" + ", ".join(_dfs) + "]"
                _args.append(_dfs)
            elif isinstance(v, (str, bool, dict, list)):
                _args.append(pformat(v, compact=True, sort_dicts=False, width=800))
            else:
                _args.append(str(v))

        _args = ', '.join(_args)
        _kwargs = []

        # Process keywords arguments
        for k, v in kwargs.items():
            if is_function(v):
                _kwargs.append(k + "=" + v.__qualname__)
            elif isinstance(v, (BaseDataFrame,)):
                _df = "self.create_dataframe(data=" + v.export(data_types="internal") + ", force_data_types=True)"
                _kwargs.append(k + "=" + _df)
            elif isinstance(v, list) and isinstance(v[0], (BaseDataFrame,)):
                _dfs = ["self.create_dataframe(data=" + dv.export(data_types="internal") + ", force_data_types=True)" for dv in v]
                _dfs = "[" + ", ".join(_dfs) + "]"
                _kwargs.append(k + "=" + _dfs)
            elif isinstance(v, (str, bool, dict, list)):
                _kwargs.append(k + "=" + pformat(v, compact=True, sort_dicts=False, width=800))
            else:
                _kwargs.append(k + "=" + str(v))

        # Separator if we have positional and keyword arguments
        separator = ""
        if (not is_list_empty(args)) & (not is_list_empty(kwargs)):
            separator = ", "

        if method is None:
            add_buffer("result = df\n", 2)

        else:
            ams = ""
            for m in additional_method:
                ams += "." + m + "()"

            add_buffer("result = df." + method + "(" + _args + separator + ', '.join(
                _kwargs) + ")" + ams + "\n", 2)

        # print("expected_df", expected_df)

        failed = False

        # Apply function
        if method is None:
            expected_df = df_func
        else:
            # Here we construct the method to be applied to the source object
            _df_func = df_func
            for f in method.split("."):
                df_func = getattr(df_func, f)

            try:
                expected_df = df_func(*args, **kwargs)
            except Exception as e:
                warnings.warn(
                    f"The operation on test creation {func_test_name} failed, passing the same dataset instead")
                print(e)
                failed = True
                expected_df = _df_func
        # Additional Methods
        for m in additional_method:
            expected_df = getattr(expected_df, m)()

        # Checks if output is ok

        expected_is_df = isinstance(expected_df, (BaseDataFrame,))

        if compare_by == "df" and not expected_is_df:
            compare_by = "dict"

        if compare_by != "df" and expected_is_df:
            add_buffer("result = result.to_dict()\n", 2)

        if failed:
            add_buffer("# The following value does not represent a correct output of the operation\n", 2)
            add_buffer("expected = self.dict\n", 2)
        elif compare_by == "df":
            if expected_is_df:
                expected_df = expected_df.export(data_types="internal")
            add_buffer(f"expected = self.create_dataframe(data={expected_df}, force_data_types=True)\n", 2)
        else:
            if expected_is_df:
                expected_df = expected_df.export(data_types=False)
                add_buffer(f"expected = {expected_df}\n", 2)
            elif is_str(expected_df):
                add_buffer(f"expected = {pformat(expected_df, compact=True, sort_dicts=False)}\n", 2)
            else:
                add_buffer(f"expected = {expected_df}\n", 2)

        # Output
        if compare_by == "df":
            add_buffer("self.assertTrue(result.equals(expected, decimal=True, assertion=True))\n", 2)
        elif compare_by == "dict":
            if expected_is_df:
                add_buffer("self.assertTrue(df_dicts_equal(result, expected, assertion=True))\n", 2)
            else:
                add_buffer("self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))\n", 2)

        elif compare_by == "json":
            add_buffer("self.assertEqual(json_encoding(result), json_encoding(expected))\n", 2)
        else:
            add_buffer("self.assertEqual(result, expected)\n", 2)

        filename = self.create_path + "/" + filename
        if not os.path.exists(os.path.dirname(filename)):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        # Write file
        test_file = open(filename, 'w', encoding='utf-8')

        for b in buffer:
            test_file.write(b)

        test_file.close()

        self.created.append((method, variant))

        # return "".join(buffer)

    def delete(self, method=None, variant=None):
        """
        This is a helper function that delete python tests files used to construct the final Test file.
        :param df: Do nothing, only for simplicity so you can delete a file test the same way you create it
        :param variant: The create method will try to create a test function with the func param given.
        If you want to test a function with different params you can use variant.
        :param method: Method to be tested
        :param method: Variant of the method
        :return:
        """

        if variant is None:
            variant = ""
        elif method is not None:
            variant = "_" + variant

        # Create func test name. If is None we just test the create.df function a not transform the data frame in
        # any way
        if method is None:
            method_test_name = "test_" + variant + "()"
            filename = variant + ".test"

        else:
            method_test_name = "test_" + \
                method.replace(".", "_") + variant + "()"

            filename = method.replace(".", "_") + variant + ".test"

        filename = self.create_path + "/" + filename

        print(f"Deleting file {filename}...")
        logger.print(method_test_name)
        try:
            os.remove(filename)
        except FileNotFoundError as e:
            warnings.warn(e)


default_configs = {
    "Pandas": {"engine": "pandas"},
    "Dask": {"engine": "dask", "n_partitions": 1},
    "PartitionDask": {"engine": "dask", "n_partitions": 2},
    "CUDF": {"engine": "cudf", "try_package": "cudf"},
    "DC": {"engine": "dask_cudf", "n_partitions": 1, "try_package": "dask_cudf"},
    "PartitionDC": {"engine": "dask_cudf", "n_partitions": 2, "try_package": "dask_cudf"},
    "Spark": {"engine": "spark", "try_package": "pyspark"},
    "Vaex": {"engine": "vaex", "try_package": "vaex"}
}



import inspect
import re
from inspect import signature
from pprint import pformat
from typing import Union, Optional

from multipledispatch.dispatcher import MethodDispatcher

from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.base.columns import BaseColumns
from optimus.engines.base.create import BaseCreate
from optimus.engines.base.engine import BaseEngine
from optimus.engines.base.stringclustering import Clusters


def engine_accessors():
    from optimus.engines.base.create import BaseCreate
    from optimus.engines.base.io.load import BaseLoad
    from optimus.engines.base.io.save import BaseSave
    from optimus.engines.base.io.connect import Connect

    return {
        "create": BaseCreate,
        "load": BaseLoad,
        "save": BaseSave,
        "connect": Connect
    }

def dataframe_accessors():
    from optimus.engines.base.columns import BaseColumns
    from optimus.engines.base.rows import BaseRows
    from optimus.engines.base.set import BaseSet
    from optimus.engines.base.mask import Mask
    from optimus.engines.base.ml.models import BaseML
    from optimus.plots.plots import Plot
    from optimus.outliers.outliers import Outliers
    from optimus.engines.base.profile import BaseProfile

    return {
        "cols": BaseColumns,
        "rows": BaseRows,
        "set": BaseSet,
        "mask": Mask,
        "ml": BaseML,
        "plot": Plot,
        "outliers": Outliers,
        "profile": BaseProfile
    }

def accessors():
    return {**engine_accessors(), **dataframe_accessors()}

def _create_new_variable(base_name: str, names: List[str]):
    while base_name in names:
        base_name = _increment_variable_name(base_name)

    return base_name


def _increment_variable_name(variable_name: str):
    match = re.search(r'\d+$', variable_name)
    if match:
        variable_name = variable_name[0:match.start()] + str(int(variable_name[match.start():match.endpos]) + 1)
    else:
        variable_name = variable_name + "2"
    return variable_name


def _arguments(args: dict, args_properties: Optional[dict] = None):
    args_list = []

    # It doesn't find properties or allows kwargs
    if not args_properties or "kwargs" in args_properties:
        args_list = list(args.keys())
        for key in ["source", "target", "operation", "operation_options"]:
            if key in args_list:
                args_list.remove(key)

    # Properties from declaration
    else:
        args_list = args_properties.keys()

    # Formating
    for arg in args_list:
        # Variable names of list of variable names
        if arg in args:
            if arg in args_properties and is_any_optimus_type(args_properties[arg].get("type", None)):
                if is_list(args[arg]):
                    args[arg] = f"[{', '.join(args[arg])}]"
                else:
                    args[arg] = args[arg]
            # Native types
            elif is_str(arg) and arg in args_properties and args_properties[arg].get("type", None) == dict:
                pass
            else:
                args[arg] = pformat(args[arg])

    return ", ".join([f'{arg}={args[arg]}' for arg in args_list if arg in args])


def _generate_code_target(body: dict, properties: dict, target: str):
    arguments = _arguments(body, properties["arguments"])

    code = ''

    if target:
        code += f'{target} = '

    if body.get("source", None):
        code += f'{body["source"]}.'

    code += f'{body["operation"]}({arguments})'

    if target:
        return code, [target]
    else:
        return code, []


def _generate_code_dataframe_transformation(body: dict, properties: dict, variables):
    target = body.get("target")

    if target is None:
        target = body["source"]

    options = body.get("operation_options", None)
    if options:
        if options.get("creates_new", False):
            target = available_variable("df", variables)
    return _generate_code_target(body, properties, target)


def _generate_code_dataframe_mask(body, properties, variables):
    target = body.get("target", available_variable("mask", variables))
    return _generate_code_target(body, properties, target)


def _generate_code_dataframe_clusters(body, properties, variables):
    target = body.get("target", available_variable("clusters", variables))
    return _generate_code_target(body, properties, target)


def _generate_code_output(body, properties, variables):
    target = body.get("target", "result")
    return _generate_code_target(body, properties, target)


def _generate_code_engine_dataframe(body, properties, variables):
    target = body.get("target", available_variable("df", variables))
    return _generate_code_target(body, properties, target)


def _generate_code_engine_connection(body, properties, variables):
    target = body.get("target", available_variable("conn", variables))
    return _generate_code_target(body, properties, target)


def _generate_code_engine(body, properties, variables):
    target = body.get("target", available_variable("op", variables))
    return _generate_code_target(body, properties, target)


def _get_generator(func_properties, method_root_type):
    if method_root_type == "dataframe":

        # If the method returns a dataframe, it's a transformation
        if 'DataFrameType' == func_properties["return_annotation"]:
            return _generate_code_dataframe_transformation
        # TO-DO: should mask functions be treated as transformations? (MaskDataFrameType)
        if 'MaskDataFrameType' == func_properties["return_annotation"]:
            return _generate_code_dataframe_mask
        # If the method returns anything else, it's an output
        elif 'ClustersType' == func_properties["return_annotation"]:
            return _generate_code_dataframe_clusters
        else:
            return _generate_code_output

    elif method_root_type == "engine":

        if 'DataFrameType' == func_properties["return_annotation"]:
            return _generate_code_engine_dataframe
        elif 'ConnectionType' == func_properties["return_annotation"]:
            return _generate_code_engine_connection
        else:
            return _generate_code_output

    elif method_root_type == "optimus":

        return _generate_code_engine

    elif method_root_type == "clusters":

        return _generate_code_output


def _get_method_root_type(accessor):
    if accessor in engine_accessors():
        return "engine"
    if accessor in dataframe_accessors():
        return "dataframe"


def _init_methods(engine):
    method = ""

    if engine == "pandas":
        from optimus.engines.pandas.engine import PandasEngine
        method = PandasEngine
    if engine == "vaex":
        from optimus.engines.vaex.engine import VaexEngine
        method = VaexEngine
    if engine == "spark":
        from optimus.engines.spark.engine import SparkEngine
        method = SparkEngine
    if engine == "dask":
        from optimus.engines.dask.engine import DaskEngine
        method = DaskEngine
    if engine == "ibis":
        from optimus.engines.ibis.engine import IbisEngine
        method = IbisEngine
    if engine == "cudf":
        from optimus.engines.cudf.engine import CUDFEngine
        method = CUDFEngine
    if engine == "dask_cudf":
        from optimus.engines.dask_cudf.engine import DaskCUDFEngine
        method = DaskCUDFEngine

    return method


def _generate_code(body=None, variables=[], **kwargs):
    if not body:
        body = kwargs

    operation = body["operation"].split(".")

    method = None
    method_root_type = None

    _accessors = accessors()
    
    if operation[0] == "Optimus":
        method = _init_methods(body["engine"])
        method_root_type = "optimus"
        operation = []
    elif operation[0] in _accessors:
        method = _accessors[operation[0]]
        method_root_type = _get_method_root_type(operation[0])
        operation = operation[1:]
    elif getattr(BaseDataFrame, operation[0], None):
        method = BaseDataFrame
        method_root_type = "dataframe"
    elif getattr(BaseEngine, operation[0], None):
        method = BaseEngine
        method_root_type = "engine"
    elif getattr(Clusters, operation[0], None):
        method = Clusters
        method_root_type = "clusters"

    for item in operation:
        method = getattr(method, item)

    properties = method_properties(method, method_root_type)

    code, updated = properties["generator"](body, properties, variables)

    return code, updated


def optimus_variables():
    from optimus.helpers.functions import list_engines, list_dataframes, list_clusters, list_connections
    return [*list_engines(), *list_dataframes(), *list_clusters(), *list_connections()]


def available_variable(name: str, variables: dict):
    return _create_new_variable(name, [*variables, *optimus_variables()])


def method_properties(func: Union[callable], method_root_type):
    try:
        fp = signature(func)
        func_properties = {"parameters": fp.parameters, "return_annotation": fp.return_annotation}
    except ValueError as e:
        if getattr(func, "funcs"):
            func_properties = {"parameters": {}, "return_annotation": None}
            for f in func.funcs.values():
                fp = signature(f)
                func_properties["parameters"].update(fp.parameters)
                if fp.return_annotation:
                    func_properties["return_annotation"] = fp.return_annotation
        else:
            raise e

    if list(func_properties["parameters"].keys()) == ['root'] and getattr(func, "__call__"):
        fp = signature(func.__call__)
        func_properties = {"parameters": fp.parameters, "return_annotation": fp.return_annotation}

    arguments_list = list(func_properties["parameters"].items())
    arguments = {}
    for key, arg in arguments_list:

        if arg.name in ['self', 'cls']:
            continue

        arguments[arg.name] = {}

        if arg.annotation is not inspect._empty:
            arguments[arg.name].update({"type": arg.annotation})

        if arg.default is not inspect._empty:
            arguments[arg.name].update({"value": arg.default})

    return {
        "arguments": arguments,
        "returns": func_properties["return_annotation"],
        "generator": _get_generator(func_properties, method_root_type)
    }


def generate_code(body: Optional[dict] = None, variables: List[str] = [], get_updated: bool = False, **kwargs):
    if not body:
        body = kwargs

    body = val_to_list(body)

    updated = []

    code = []

    for operation in body:
        operation = prepare(operation)
        operation_code, operation_updated = _generate_code(operation, [*updated, *variables])
        updated.extend(operation_updated)
        code.append(operation_code)

    if get_updated:
        return "\n".join(code), one_list_to_val(updated)
    else:
        return "\n".join(code)



import math

from pyspark.ml.feature import SQLTransformer
from pyspark.serializers import AutoBatchedSerializer, PickleSerializer
from pyspark.sql import functions as F

from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import random_int
from optimus.helpers.functions_spark import traverse
from optimus.helpers.raiseit import RaiseIt
from optimus.helpers.types import *
from optimus.infer import is_str


class SparkDataFrame(BaseDataFrame):
    def _base_to_dfd(self, pdf, n_partitions) -> 'InternalDataFrameType':
        pass


    # Koalas do not support the 'or' operation over the whole dataframe.
    # def __or__(self, df2) -> 'DataFrameType':
    #     return self.new(((self.get_series()) | (df2.get_series())).to_frame())

    def visualize(self):
        pass

    def graph(self) -> dict:
        pass

    @property
    def rows(self):
        from optimus.engines.spark.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.spark.columns import Cols
        return Cols(self)

    @property
    def save(self):
        from optimus.engines.spark.io.save import Save
        return Save(self)

    @property
    def constants(self):
        from optimus.engines.spark.constants import Constants
        return Constants()

    @property
    def encoding(self):
        from optimus.engines.spark.ml.encoding import Encoding
        return Encoding(self)

    @property
    def functions(self):
        from optimus.engines.spark.functions import SparkFunctions
        return SparkFunctions(self)

    def _iloc(self, lower_bound, upper_bound, copy=True):
        dfd = self.data[lower_bound: upper_bound]
        if copy:
            dfd = dfd.reset_index(drop=True)
        return self.__class__(dfd, op=self.op, label_encoder=self.le, meta=self.meta)

    def execute(self):
        df = self.data
        return self.new(df.cache(), meta=self.meta)


    def export(self, n="all", data_types="inferred"):
        """
        Helper function to export all the dataframe in text format. Aimed to be used in test functions
        :return:
        """
        df = self.root
        dfd = df.data
        dict_result = {}
        value = dfd.collect()
        schema = []
        for col_names in df.cols.names():
            name = col_names

            data_type = df.cols.schema_data_type(col_names)
            if isinstance(data_type, ArrayType):
                data_type = "ArrayType(" + str(data_type.elementType) + "()," + str(data_type.containsNull) + ")"
            else:
                data_type = str(data_type) + "()"

            nullable = df.schema[col_names].nullable

            schema.append(
                "('{name}', {dataType}, {nullable})".format(name=name, dataType=data_type, nullable=nullable))
        schema = ",".join(schema)
        schema = "[" + schema + "]"

        # if there is only an element in the dict just return the value
        if len(dict_result) == 1:
            dict_result = next(iter(dict_result.values()))
        else:
            dict_result = [tuple(v.asDict().values()) for v in value]

        def func(path, _value):
            try:
                if math.isnan(_value):
                    r = None
                else:
                    r = _value
            except TypeError:
                r = _value
            return r

        dict_result = traverse(dict_result, None, func)

        return "{schema}, {dict_result}".format(schema=schema, dict_result=dict_result)


    def sample(self, n=10, random=False):
        """
        Return a n number of sample from a dataFrame
        :param n: Number of samples
        :param random: if true get a semi random sample
        :return:
        """
        df = self.root
        if random is True:
            seed = random_int()
        elif random is False:
            seed = 0
        else:
            RaiseIt.value_error(random, ["True", "False"])

        rows_count = df.rows.count()
        if n < rows_count:
            # n/rows_count can return a number that represent less the total number we expect. multiply by 1.1 bo
            fraction = (n / rows_count) * 1.1
        else:
            fraction = 1.0

        return self.new(self.data.sample(False, fraction, seed=seed).limit(n))

    def stratified_sample(self, col_name, seed: int = 1):
        """
        Stratified Sampling
        :param col_name:
        :param seed:
        :return:
        """
        df = self.root.data.to_spark()
        fractions = df.select(col_name).distinct().withColumn("fraction", F.lit(0.8)).rdd.collectAsMap()
        df = df.stat.sampleBy(col_name, fractions, seed).to_koalas()
        return self.new(df)

    def join(self, df_right: 'DataFrameType', how="left", on=None, left_on=None, right_on=None,
             key_middle=False) -> 'DataFrameType':
        """

        :param df_right:
        :param how:
        :param on:
        :param left_on:
        :param right_on:
        :param key_middle:
        :return:
        """

        suffix_left = "_left"
        suffix_right = "_right"

        if on is not None:
            left_on = on
            right_on = on
        suffix_left_name = left_on + suffix_left
        suffix_right_name = right_on + suffix_right

        df_left = self.root.data.to_spark().withColumnRenamed(left_on, suffix_left_name)
        df_right = df_right.data.to_spark().withColumnRenamed(left_on, suffix_right_name)

        return self.new(df_left.join(df_right, df_left[left_on + suffix_left] == df_right[right_on + suffix_right],
                                     how).to_koalas())

    def pivot(self, col, groupby, agg=None, values=None):
        """
        Return reshaped DataFrame organized by given index / column values.
        :param self: Spark Dataframe
        :param col: Column to use to make new frame's columns.
        :param values: Column(s) to use for populating new frame's values.
        :return:
        """
        if values is not None:
            agg = (F.first(values))

        groupby = val_to_list(groupby)
        if agg is None:
            agg = (F.count(col))

        dfd = self.root.data
        return self.new(dfd.to_spark().groupby(groupby).pivot(col).agg(agg).to_koalas())

    @staticmethod
    def melt(id_vars, value_vars, var_name="variable", value_name="value", data_type="str"):
        """
        Convert DataFrame from wide to long format.
        :param id_vars: column with unique values
        :param value_vars: Column names that are going to be converted to columns values
        :param var_name: Column name for vars
        :param value_name: Column name for values
        :param data_type: All columns must have the same type. It will transform all columns to this data type.
        :return:
        """

        df = self
        id_vars = val_to_list(id_vars)
        # Cast all columns to the same type
        df = df.cols.cast(id_vars + value_vars, data_type)

        vars_and_vals = [F.struct(F.lit(c).alias(var_name), F.col(c).alias(value_name)) for c in value_vars]

        # Add to the DataFrame and explode
        df = df.withColumn("vars_and_vals", F.explode(F.array(*vars_and_vals)))

        cols = id_vars + [F.col("vars_and_vals")[x].alias(x) for x in [var_name, value_name]]

        return df.select(*cols)

    @staticmethod
    def size():
        """
        Get the size of a dataframe in bytes
        :return:
        """

        def _to_java_object_rdd(rdd):
            """
            Return a JavaRDD of Object by unpickling
            It will convert each Python object into Java object by Pyrolite, whenever the
            RDD is serialized in batch or not.
            """
            rdd = rdd._reserialize(AutoBatchedSerializer(PickleSerializer()))
            return rdd.ctx._jvm.org.apache.spark.mllib.api.python.SerDe.pythonToJava(rdd._jrdd, True)

        if version.parse(Spark.instance.spark.version) < version.parse("2.4.0"):
            java_obj = _to_java_object_rdd(self.rdd)
            n_bytes = Spark.instance.sc._jvm.org.apache.spark.util.SizeEstimator.estimate(java_obj)
        else:
            # TODO: Find a way to calculate the dataframe size in spark 2.4
            n_bytes = -1

        return n_bytes


    def compute(self):
        """
        This method is a very useful function to break lineage of transformations. By default Spark uses the lazy
        evaluation approach in processing data: transformation functions are not computed into an action is called.
        Sometimes when transformations are numerous, the computations are very extensive because the high number of
        operations that spark needs to run in order to get the results.

        Other important thing is that Apache Spark save task but not result of dataFrame, so tasks are
        accumulated and the same situation happens.

        :param self: Spark Dataframe
        :return:
        """

        self.data.cache().count()
        return self


    def query(self, sql_expression):
        """
        Implements the transformations which are defined by SQL statement. Currently we only support
        SQL syntax like "SELECT ... FROM __THIS__ ..." where "__THIS__" represents the
        underlying table of the input dataframe.
        :param self: Spark Dataframe
        :param sql_expression: SQL expression.
        :return: Dataframe with columns changed by SQL statement.
        """
        dfd = self.root.data
        sql_transformer = SQLTransformer(statement=sql_expression)
        return sql_transformer.transform(dfd)


    def set_name(self, value=None):
        """
        Create a temp view for a data frame also used in the json output profiling
        :param value:
        :return:
        """
        dfd = self.root.data
        dfd._name = value
        if not is_str(value):
            RaiseIt.type_error(value, ["string"])

        if len(value) == 0:
            RaiseIt.value_error(value, ["> 0"])

        dfd.createOrReplaceTempView(value)

    def partitions(self):
        """
        Return the dataframe partitions number
        :return: Number of partitions
        """
        df = self.data

        return df.to_spark().rdd.getNumPartitions()


    def partitioner(self):
        """
        Return the algorithm used to partition the dataframe
        :param self: Spark Dataframe
        :return:
        """
        dfd = self.root.data
        return dfd.rdd.partitioner

    def repartition(self, n=None, *args, **kwargs):
        df = self.data
        df = df.repartition(n, *args, **kwargs)
        return self.new(df, meta=self.meta)

    def h_repartition(self, partitions_number=None, col_name=None):
        """
        Apply a repartition to a datataframe based in some heuristics. Also you can pass the number of partitions and
        a column if you need more control.
        See
        https://stackoverflow.com/questions/35800795/number-of-partitions-in-rdd-and-performance-in-spark/35804407#35804407
        https://medium.com/@adrianchang/apache-spark-partitioning-e9faab369d14
        https://umbertogriffo.gitbooks.io/apache-spark-best-practices-and-tuning/content/sparksqlshufflepartitions_draft.html
        :param self: Spark Dataframe
        :param partitions_number: Number of partitions
        :param col_name: Column to be used to apply the repartition id necessary
        :return:
        """

        if partitions_number is None:
            partitions_number = Spark.instance.parallelism * 4

        if col_name is None:
            df = self.repartition(partitions_number)
        else:
            df = self.repartition(partitions_number, col_name)
        return df

    def show(self, n=10):
        """
        :return:
        """
        self.data.show(n)


    def random_split(self, weights: list = None, seed: int = 1):
        """
        Create 2 random splited DataFrames
        :param weights:
        :param seed:
        :return:
        """
        dfd = self.root.data
        if weights is None:
            weights = [0.8, 0.2]

        return dfd.randomSplit(weights, seed)

    def debug(self):
        """
        :return:
        """
        df = self.data
        print(df.rdd.toDebugString().decode("ascii"))

    def to_optimus_pandas(self):
        return PandasDataFrame(self.data.toPandas(), op=self.op)

    def to_pandas(self):
        return self.data.toPandas()



import glob
import ntpath
import os
from abc import abstractmethod

import psutil

from optimus.engines.base.meta import Meta
from optimus.engines.pandas.ml.models import Model
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger
from optimus.helpers.raiseit import RaiseIt
from optimus.helpers.types import DataFrameType, InternalDataFrameType
from optimus.infer import is_empty_function, is_list, is_str, is_url

XML_THRESHOLD = 10
JSON_THRESHOLD = 20
BYTES_SIZE = 1310720


class BaseLoad:

    def __init__(self, op):
        self.op = op

    @abstractmethod
    def df(self, *args, **kwargs) -> 'DataFrameType':
        pass

    def _csv(self, *args, **kwargs) -> 'InternalDataFrameType':
        pass

    def _json(self, *args, **kwargs) -> 'InternalDataFrameType':
        pass

    def _excel(self, *args, **kwargs) -> 'InternalDataFrameType':
        pass

    def _avro(self, *args, **kwargs) -> 'InternalDataFrameType':
        pass

    def _xml(self, *args, **kwargs) -> 'InternalDataFrameType':
        pass

    def _parquet(self, *args, **kwargs) -> 'InternalDataFrameType':
        pass

    def csv(self, filepath_or_buffer, sep=",", header=True, infer_schema=True, encoding="UTF-8", n_rows=None,
            null_value="None", quoting=3, lineterminator='\r\n', on_bad_lines='warn', cache=False, na_filter=False,
            storage_options=None, conn=None, *args, **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a csv file. It is the same read.csv Spark function with some predefined
        params.

        :param filepath_or_buffer: Path or location of the file.
        :param sep: Usually delimiter mark are ',' or ';'.
        :param encoding: Encoding to use for UTF when reading/writing.
        :param header: Tell the function whether dataset has a header row. True default.
        :param infer_schema: Infers the input schema automatically from data.
        :param n_rows:
        :param null_value: Whether or not to include the default NaN values when parsing the data
        :param quoting: Control field quoting behavior
        :param cache:
        :param na_filter: Detect missing value markers (empty strings and the value of na_values).
        :param lineterminator: Character to break file into lines.
        :param on_bad_lines: Specifies what to do upon encountering a bad line (a line with too many fields).
            Allowed values are:
            ‘error’, raise an Exception when a bad line is encountered.
            ‘warn’, raise a warning when a bad line is encountered and skip that line.
            ‘skip’, skip bad lines without raising or warning when they are encountered
        :param storage_options: A dict with the connection params.
        :param conn: A connection object.
        It requires one extra pass over the data. True default.

        :return dataFrame
        """

        if is_empty_function(self._csv):
            raise NotImplementedError(f"'load.csv' is not implemented on '{self.op.engine_label}'")

        unquoted_path = None

        if not is_url(filepath_or_buffer):
            unquoted_path = glob.glob(unquote_path(filepath_or_buffer))

        if unquoted_path and len(unquoted_path):
            meta = {"file_name": unquoted_path[0], "name": ntpath.basename(unquoted_path[0])}
        else:
            meta = {"file_name": filepath_or_buffer, "name": ntpath.basename(filepath_or_buffer)}

        try:

            # Pandas do not support \r\n terminator.
            if lineterminator and lineterminator.encode(encoding='UTF-8', errors='strict') == b'\r\n':
                lineterminator = None

            if conn is not None:
                filepath_or_buffer = [conn.path(fb) for fb in val_to_list(filepath_or_buffer)]
                storage_options = conn.storage_options

            if kwargs.get("chunk_size") == "auto":
                # Chunk size is going to be 75% of the memory available
                kwargs.pop("chunk_size")
                kwargs["chunksize"] = psutil.virtual_memory().free * 0.75

            na_filter = na_filter if null_value else False

            if not is_str(on_bad_lines):
                on_bad_lines = 'error' if on_bad_lines else 'skip'

            def _read(_filepath_or_buffer):
                return self._csv(_filepath_or_buffer, sep=sep, header=0 if header else None, encoding=encoding,
                                 nrows=n_rows, quoting=quoting, lineterminator=lineterminator,
                                 on_bad_lines=on_bad_lines, na_filter=na_filter,
                                 na_values=val_to_list(null_value), index_col=False,
                                 storage_options=storage_options, *args, **kwargs)

            if is_list(filepath_or_buffer):
                df = self.op.F.new_df()
                for f in filepath_or_buffer:
                    df = df.append(_read(f))
            else:
                df = _read(filepath_or_buffer)

            df = self.df(df, op=self.op)

            df.meta = Meta.set(df.meta, value=meta)

        except IOError as error:
            logger.print(error)
            raise

        return df

    def xml(self, path, n_rows=None, storage_options=None, conn=None, *args, **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a XML file.

        :param path: Any valid XML string or path is acceptable. The string could be a URL.
        :param n_rows:
        :param storage_options: A dict with the connection params.
        :param conn: A connection object.
        :param args:
        :param kwargs:
        :return:
        """

        if is_empty_function(self._xml):
            raise NotImplementedError(f"'load.xml' is not implemented on '{self.op.engine_label}'")

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "xml")[0]

        try:
            df = self._xml(file, n_rows, storage_options=storage_options, *args, **kwargs)
            df = self.df(df, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        except IOError as error:
            logger.print(error)
            raise

        return df

    def json(self, filepath_or_buffer, multiline=False, n_rows=False, storage_options=None,
             conn=None, *args, **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a json file.

        :param filepath_or_buffer: path or location of the file.
        :param multiline:
        :param n_rows:
        :param storage_options: A dict with the connection params.
        :param conn: A connection object.
        :param args:
        :param kwargs:
        :return:
        """

        if is_empty_function(self._json):
            raise NotImplementedError(f"'load.json' is not implemented on '{self.op.engine_label}'")

        if conn is not None:
            filepath_or_buffer = conn.path(filepath_or_buffer)
            storage_options = conn.storage_options

        if n_rows:
            kwargs["nrows"] = n_rows
            multiline = True

        if is_str(filepath_or_buffer):
            try:
                filepath_or_buffer = unquote_path(filepath_or_buffer)
                local_file_names = prepare_path(filepath_or_buffer, "json")
                df_list = []

                for file_name, j in local_file_names:
                    df = self._json(file_name, lines=multiline, *args, **kwargs)
                    df_list.append(df)
                df = self.op.F.df_concat(df_list)
                df = self.df(df, op=self.op)

                file_name = local_file_names[0][1]
                df.meta = Meta.set(df.meta, "file_name", file_name)

            except IOError as error:
                logger.print(error)
                raise

        else:
            df = self._json(filepath_or_buffer, lines=multiline, storage_options=storage_options, *args, **kwargs)
            df = self.df(df, op=self.op)

        return df

    def excel(self, filepath_or_buffer, header=0, sheet_name=0, merge_sheets=False, skip_rows=0, n_rows=None,
              storage_options=None, conn=None, *args, **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a excel file.
        
        :param filepath_or_buffer: Path or location of the file. Must be string dataType
        :param header: 
        :param sheet_name: excel sheet name
        :param merge_sheets: 
        :param skip_rows: 
        :param n_rows: 
        :param storage_options: A dict with the connection params.
        :param conn: A connection object.
        :param args: custom argument to be passed to the excel function
        :param kwargs: custom keyword arguments to be passed to the excel function
        :return: 
        """

        if is_empty_function(self._excel):
            raise NotImplementedError(f"'load.excel' is not implemented on '{self.op.engine_label}'")

        if conn is not None:
            filepath_or_buffer = conn.path(filepath_or_buffer)
            storage_options = conn.storage_options

        filepath_or_buffer = unquote_path(filepath_or_buffer)
        local_file_names = prepare_path(filepath_or_buffer, "xls")

        if merge_sheets is True:
            skip_rows = -1

        file = local_file_names[0][0]

        df, sheet_names = self._excel(file, sheet_name=sheet_name, skiprows=skip_rows, header=header, nrows=n_rows,
                                      storage_options=storage_options, *args, **kwargs)

        df = self.df(df, op=self.op)

        file_name = local_file_names[0][1]
        print(local_file_names, file, file_name, filepath_or_buffer)
        df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        df.meta = Meta.set(df.meta, "sheet_names", sheet_names)

        return df

    def avro(self, filepath_or_buffer, n_rows=None, storage_options=None, conn=None,
             *args, **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a avro file.

        :param filepath_or_buffer: path or location of the file. Must be string dataType
        :param n_rows:
        :param storage_options: A dict with the connection params.
        :param conn: A connection object.
        :param args: custom argument to be passed to the spark avro function
        :param kwargs: custom keyword arguments to be passed to the spark avro function
        """

        if is_empty_function(self._avro):
            raise NotImplementedError(f"'load.avro' is not implemented on '{self.op.engine_label}'")

        filepath_or_buffer = unquote_path(filepath_or_buffer)

        if conn is not None:
            logger.warn("'load.avro' does not support connection options ('conn')")

        if storage_options is not None:
            logger.warn("'load.avro' does not support 'storage_options'")

        file, file_name = prepare_path(filepath_or_buffer, "avro")[0]

        try:
            df = self._avro(filepath_or_buffer, nrows=n_rows, *args, **kwargs)
            df = self.df(df, op=self.op)
            df.meta = Meta.set(df.meta, value={"file_name": file_name, "name": ntpath.basename(filepath_or_buffer)})

        except IOError as error:
            logger.print(error)
            raise

        return df

    def parquet(self, filepath_or_buffer, columns=None, n_rows=None, storage_options=None, conn=None,
                *args, **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a parquet file.

        :param filepath_or_buffer: path or location of the file. Must be string dataType
        :param columns: select the columns that will be loaded. In this way you do not need to load all the dataframe
        :param n_rows: number of rows to load
        :param storage_options: A dict with the connection params.
        :param conn: A connection object.
        :param args: custom argument to be passed to the spark parquet function
        :param kwargs: custom keyword arguments to be passed to the spark parquet function
        """

        if is_empty_function(self._parquet):
            raise NotImplementedError(f"'load.parquet' is not implemented on '{self.op.engine_label}'")

        filepath_or_buffer = unquote_path(filepath_or_buffer)

        if conn is not None:
            filepath_or_buffer = conn.path(filepath_or_buffer)
            storage_options = conn.storage_options

        try:
            dfd = self._parquet(filepath_or_buffer, columns=columns, nrows=n_rows,
                                storage_options=storage_options, *args, **kwargs)
            df = self.df(dfd, op=self.op)
            df.meta = Meta.set(df.meta,
                               value={"file_name": filepath_or_buffer, "name": ntpath.basename(filepath_or_buffer)})

        except IOError as error:
            logger.print(error)
            raise

        return df

    def orc(self, path, columns, storage_options=None, conn=None, n_partitions=None, *args,
            **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a OCR file.

        :param path: path or location of the file. Must be string dataType.
        :param columns: Specific column names to be loaded from the file.
        :param storage_options: A dict with the connection params.
        :param conn: A connection object.
        :param args: custom argument to be passed to the spark avro function.
        :param kwargs: custom keyword arguments to be passed to the spark avro function.
        """
        raise NotImplementedError('Not implemented yet')

    def zip(self, path, filename, dest=None, columns=None, storage_options=None, conn=None, n_partitions=None,
            *args, **kwargs) -> 'DataFrameType':
        pass

    def hdf5(self, path, columns=None, n_partitions=None, *args, **kwargs) -> 'DataFrameType':
        """
        Loads a dataframe from a HDF5 file.

        :param path: path or location of the file. Must be string dataType.
        :param columns: Specific column names to be loaded from the file.
        :param n_partitions:
        :param args: custom argument to be passed to the spark avro function.
        :param kwargs: custom keyword arguments to be passed to the spark avro function.
        :return:
        """
        raise NotImplementedError('Not implemented yet')

    def tsv(self, filepath_or_buffer, header=True, infer_schema=True, *args, **kwargs):
        """
        Loads a dataframe from a tsv(Tabular separated values) file.

        :param filepath_or_buffer: Path or location of the file. Must be string dataType
        :param header:
        :param infer_schema:
        :param args: custom argument to be passed to the spark avro function.
        :param kwargs: custom keyword arguments to be passed to the spark avro function.
        :return:
        """
        return self.csv(filepath_or_buffer, sep='\t', header=header, infer_schema=infer_schema, *args, **kwargs)

    def file(self, path, *args, **kwargs) -> 'DataFrameType':
        """
        Try to infer the file data format and encoding and load the data into a dataframe.

        :param path: Path to the file you want to load.
        :param args: custom argument to be passed to the spark avro function.
        :param kwargs: custom keyword arguments to be passed to the spark avro function.
        :return:
        """
        conn = kwargs.get("conn")

        if conn:
            import boto3
            remote_obj = boto3.resource(
                conn.type, **conn.boto).Object(conn.options.get("bucket"), path)
            body = remote_obj.get()['Body']
            buffer = body.read(amt=BYTES_SIZE)
            full_path = conn.path(path)
            file_name = os.path.basename(path)

        else:

            full_path, file_name = prepare_path(path)[0]
            file = open(full_path, "rb")
            buffer = file.read(BYTES_SIZE)

        ext_to_type = {
            "xls": "excel",
            "xlsx": "excel"
        }

        # Detect the file type
        try:
            file_ext = os.path.splitext(file_name)[1].replace(".", "")
            import magic
            mime, encoding = magic.Magic(
                mime=True, mime_encoding=True).from_buffer(buffer).split(";")
            mime_info = {"mime": mime, "encoding": encoding.strip().split("=")[
                1], "file_ext": file_ext}

        except Exception as e:
            print(getattr(e, 'message', repr(e)))
            full_path = path
            file_name = path.split('/')[-1]
            file_ext = file_name.split('.')[-1]
            mime = False
            mime_info = {"file_type": ext_to_type.get(file_ext, file_ext), "encoding": False}

        file_type = ext_to_type.get(file_ext, file_ext)

        if mime:
            if mime in ["text/plain", "application/csv"]:
                if mime_info["file_ext"] == "json":
                    file_type = "json"
                else:
                    file_type = "csv"
            elif mime == "application/json":
                file_type = "json"
            elif mime == "text/xml":
                file_type = "xml"
            elif mime in ["application/vnd.ms-excel",
                          "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"]:
                file_type = "excel"
            elif not file_type:
                RaiseIt.value_error(
                    mime, ["csv", "json", "xml", "xls", "xlsx"])

        # Detect the file encoding
        if file_type == "csv":
            # In some case magic get a "unknown-8bit" which can not be use to decode the file use latin-1 instead
            if mime_info.get("encoding", None) == "unknown-8bit":
                mime_info["encoding"] = "latin-1"

            if mime:
                import csv
                dialect = csv.Sniffer().sniff(str(buffer))
                mime_info["file_type"] = "csv"

                properties = {"sep": dialect.delimiter,
                              "doublequote": dialect.doublequote,
                              "escapechar": dialect.escapechar,
                              "lineterminator": dialect.lineterminator,
                              "quotechar": dialect.quotechar,
                              "quoting": dialect.quoting,
                              "skipinitialspace": dialect.skipinitialspace}

                for k in properties:
                    if k not in kwargs:
                        kwargs.update({k: properties[k]})

                mime_info.update({"properties": properties})
                kwargs.update({"encoding": mime_info.get("encoding", None)})

            df = self.csv(full_path, *args, **kwargs)

        elif file_type == "json":
            mime_info["file_type"] = "json"
            df = self.json(full_path, *args, **kwargs)

        elif file_type == "xml":
            mime_info["file_type"] = "xml"
            df = self.xml(full_path, **kwargs)

        elif file_type == "excel":
            mime_info["file_type"] = "excel"
            df = self.excel(full_path, **kwargs)

        elif file_type == "parquet":
            mime_info["file_type"] = "parquet"
            df = self.parquet(full_path, **kwargs)

        else:
            RaiseIt.value_error(
                file_type, ["csv", "json", "xml", "excel","parquet"])

        if file_name:
            df.meta = Meta.set(df.meta, "file_name", file_name)

        return df

    def model(self, path):
        """
        Load a machine learning model from a file.

        :param path: Path to the file we want to load.
        :return:
        """
        import joblib
        return Model(model=joblib.load(path), op=self.op)



from abc import ABC, abstractmethod


class AbstractOutlierThreshold(ABC):
    """
     This is a template class to expand the outliers methods
     Also you need to add the function to outliers.py
     """

    def select(self):
        """
        Select outliers rows using the selected column
        :return:
        """
        df = self.df
        z_score = self.z_score

        return df.rows.select(
            (z_score > self.threshold) | (z_score < (self.threshold) * -1))

    def select_lower_bound(self):
        df = self.df
        z_score = self.z_score
        return df.rows.select((z_score < (self.threshold) * -1))

    def select_upper_bound(self):
        df = self.df
        z_score = self.z_score
        return df.rows.select((z_score > (self.threshold)))

    def drop(self):
        """
        Drop outliers rows using the selected column
        :return:
        """

        df = self.df
        z_score = self.z_score
        return df.rows.drop(z_score.cols.abs() >= self.threshold)

    def count_lower_bound(self, bound: int):
        """
        Count outlier in the lower bound
        :return:
        """
        col_name = self.col_name
        return self.df.rows.select(self.df[col_name] < bound).count()

    def count_upper_bound(self, bound: int):
        """
        Count outliers in the upper bound
        :return:
        """
        col_name = self.col_name
        return self.df.rows.select(self.df[col_name] >= bound).count()

    def count(self):
        """
        Count the outliers rows using the selected column
        :return:
        """
        df = self.z_score
        return df.rows.select(df > self.threshold).rows.count(compute=False)

    def non_outliers_count(self):
        """
        Count non outliers rows using the selected column
        :return:
        """
        df = self.df
        z_score = self.z_score

        return df.rows.select(z_score < self.threshold).rows.count(compute=False)

    @abstractmethod
    def info(self, output: str = "dict"):
        """
        Get whiskers, iqrs and outliers and non outliers count
        :return:
        """
        pass
